import os
from argparse import ArgumentParser, Namespace
from configparser import ConfigParser
from typing import Dict, Optional, Sequence, Union


def config_to_dict(config: ConfigParser) -> Dict[str, Dict[str, str]]:
    content = {}
    for section in config.sections():
        if section not in content:
            content[section] = {}
        for key in config[section]:
            content[section][key] = config[section][key]
    return content


def sort_dict(
    content: Dict[str, Union[Dict[str, str], str]]
) -> Dict[str, Union[Dict[str, str], str]]:
    temp = {}
    for key in sorted(content):
        if isinstance(content[key], dict):
            temp[key] = sort_dict(content[key])
        else:
            temp[key] = content[key]
    return temp


def dict_to_config(content: Dict[str, Dict[str, str]]) -> ConfigParser:
    config = ConfigParser()
    for section, keys in content.items():
        config.add_section(section)
        for key, value in keys.items():
            config.set(section, key, value)
    return config


# https://stackoverflow.com/a/10289740
def remove_extra_line(filename):
    with open(filename, "r+", encoding="utf-8") as stream:
        stream.seek(0, os.SEEK_END)

        # This code means the following code skips the very last character in the file -
        # i.e. in the case the last line is null we delete the last line
        # and the penultimate one
        pos = stream.tell() - 1

        # Read each character in the file one at a time from the penultimate
        # character going backwards, searching for a newline character
        # If we find a new line, exit the search
        while pos > 0 and stream.read(1) != "\n":
            pos -= 1
            stream.seek(pos, os.SEEK_SET)

        # So long as we're not at the start of the file, delete all the characters ahead
        # of this position
        if pos > 0:
            stream.seek(pos + 1, os.SEEK_SET)
            stream.truncate()


def main(argv: Optional[Sequence[str]] = None) -> int:
    args = parse_arguments(argv)

    return_value = 0
    for filename in args.filenames:
        config = ConfigParser()
        config.read(filename)
        content = config_to_dict(config)
        if not args.no_sort:
            content = sort_dict(content)
        config = dict_to_config(content)
        with open(filename, "w") as stream:
            config.write(stream)
        remove_extra_line(filename)
    return return_value


def parse_arguments(args: Optional[Sequence[str]] = None) -> Namespace:
    parser = ArgumentParser()
    parser.add_argument("--no-sort", action="store_true", help="Don't sort the result")
    parser.add_argument("filenames", nargs="*", help="Filenames to format.")
    return parser.parse_args(args)


if __name__ == "__main__":
    raise SystemExit(main())
