from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Optional, Sequence, Union

from ruamel.yaml import YAML, CommentedMap, CommentedSeq, YAMLError


def sort_data(content: Union[CommentedMap, CommentedSeq]) -> Union[CommentedMap, CommentedSeq]:
    if isinstance(content, CommentedMap):
        return sort_dict(content=content)
    elif isinstance(content, CommentedSeq):
        return sort_list(content=content)
    return content


def sort_dict(content: CommentedMap) -> CommentedMap:
    for key in sorted(content, reverse=True):
        value = content.pop(key)
        if isinstance(value, CommentedMap):
            content.insert(0, key, sort_dict(content=value))
        elif isinstance(value, CommentedSeq):
            content.insert(0, key, sort_list(content=value))
        else:
            content.insert(0, key, value)
    repo_start_keys = ["rev", "repo"]
    for key in repo_start_keys:
        if key in content:
            content.insert(0, key, content.pop(key))
    if "hooks" in content:
        content["hooks"] = content.pop("hooks")
    hook_start_keys = ["name", "alias", "id"]
    for key in hook_start_keys:
        if key in content:
            content.insert(0, key, content.pop(key))
    return content


def sort_list(content: CommentedSeq) -> CommentedSeq:
    for index, value in enumerate(content):
        if isinstance(value, CommentedMap):
            content[index] = sort_dict(content=value)
        elif isinstance(value, CommentedSeq):
            content[index] = sort_list(content=value)
    if content:
        if "repo" in content[0]:
            content.sort(key=lambda x: x["repo"].lower())
        elif "id" in content[0]:
            content.sort(
                key=lambda x: (
                    x["id"].lower(),
                    x["name"].lower() if "name" in x else "",
                    x["alias"].lower() if "alias" in x else "",
                )
            )
        elif not isinstance(content[0], CommentedMap):
            content.sort()
    return content


def style_content(content: Union[CommentedMap, CommentedSeq], style: Optional[str] = None):
    if isinstance(content, CommentedMap):
        for key in content:
            if isinstance(content[key], CommentedMap) or isinstance(content[key], CommentedSeq):
                style_content(content=content[key], style=style)
    elif isinstance(content, CommentedSeq):
        for entry in content:
            if isinstance(entry, CommentedMap) or isinstance(entry, CommentedSeq):
                style_content(content=entry, style=style)
    if style == "flow":
        content.fa.set_flow_style()
    elif style == "block":
        content.fa.set_block_style()


def yaml_setup(line_length: int = 100, indent: int = 2) -> YAML:
    def null_representer(self, data):
        return self.represent_scalar("tag:yaml.org,2002:null", "~")

    yaml = YAML(pure=True)
    yaml.default_flow_style = False
    yaml.width = line_length
    yaml.indent(mapping=indent, sequence=indent + 2, offset=indent)
    yaml.representer.add_representer(type(None), null_representer)
    # yaml.emitter.alt_null = '~'
    return yaml


def main(argv: Optional[Sequence[str]] = None) -> int:
    args = parse_arguments(argv)

    return_value = 0
    yaml = yaml_setup(line_length=args.line_length, indent=args.indent)
    for filename in args.filenames:
        yaml_file = Path(filename).resolve()
        with yaml_file.open("r") as stream:
            try:
                content = yaml.load(stream)
            except YAMLError as err:
                print(f"Unable to parse {filename}: {err}")
                return_value = 1
                continue
        style_content(content=content, style="block")
        if not args.no_sort:
            content = sort_data(content=content)
        with yaml_file.open("w") as stream:
            yaml.dump(content, stream)
    return return_value


def parse_arguments(args: Optional[Sequence[str]] = None) -> Namespace:
    parser = ArgumentParser()
    parser.add_argument(
        "--line-length",
        type=int,
        default=100,
        help="The max line length before wrapping (Default: 100)",
    )
    parser.add_argument(
        "--indent",
        type=int,
        default=2,
        help="The number of spaces for indentation (Default: 2)",
    )
    parser.add_argument("--no-sort", action="store_true", help="Don't sort the result")
    parser.add_argument("filenames", nargs="*", help="Filenames to format.")
    return parser.parse_args(args)


if __name__ == "__main__":
    raise SystemExit(main())
